/*********************************************************************
 * Photon transport routines, shared between parallel and non-parallel
 *  versions of the code.  Propagate a single photon until it either
 *  exists the system or runs out of time.
 *********************************************************************/

/*
 * This file is part of tMCimg.
 * 
 * tMCimg is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "config.h"
#include "tMCimg.h"
#include "mccore.h"

#if (HAVE_ASSERT_H)
#include <assert.h>
#endif

#include "commonMC.h"

/*********************************************************************
 Function prototypes
 *********************************************************************/

void run_photon(const struct Config *, int, const TISSUE ***, 
                TWOPT_T *, TWOPT_T *, FILE *, int *);

static void renorm(MCCOORD *);
static void rotate_phase(TWOPT_T *, TWOPT_T *, 
                         const struct Config *, const float *);
static void scatter(double, double *, double *, double *, int *);
static int reflect_photon(const struct Config *, const MCCOORD *, 
                          const MCCOORD *, double *, double *, double *, 
                          int *, const TISSUE ***);
static void find_interface(const MCCOORD *, MCCOORD *,
                           const struct Config *, const TISSUE ***);
static void score_photon(const struct Config *, const TISSUE ***, 
                         const MCCOORD *, double, TWOPT_T, TWOPT_T,
                         TWOPT_T *, TWOPT_T *);

extern void write_photon(const struct Config *, FILE *,
			 double, double, double,
			 double, double, double, double, float *);
static int fiberNA(double, double);

/**************************************************************
  RUN A SINGLE PHOTON THROUGH THE SIMULATION
**************************************************************/

void run_photon(const struct Config *syscfg, int ns,
		const TISSUE ***tissueType, TWOPT_T *II, TWOPT_T *JJ,
		FILE *fp, int *seed)
{
  double  x, y, z;	      /* INITIAL PHOTON POSITION */
  MCCOORD R, Rnew;	      /* CURRENT PHOTON POSITION */
  double  kx, ky, kz;	      /* DIRECTION COSINES */
  double  Lresid, Tresid, Ttot, maxstep, step;
  TWOPT_T P2pt, Q2pt;	      /* PHOTON WEIGHT */
  float   lenTiss[MAXTISSUE]; /* LENGTH SPENT IN EACH TISSUE */
  double  rnm;		      /* RANDOM NUMBER */
  int i,  tissueIndex;
  double  stepT, stepL;

  /* SET THE PHOTON WEIGHT TO 1 AND INITIALIZE PHOTON LENGTH PARAMETERS */

  P2pt = 1.0;
  Q2pt = 0.0;

  Ttot   = 0.0;
  Lresid = 0.0;
  Tresid = 0.0;

  /* Seed RNG, if necessary */

  ran(seed);

  /* Initialize the length in each tissue type */

  for (i = 0; i <= MAXTISSUE; i++)
    lenTiss[i] = 0.0;

  /* Initial source position and direction */

  x = syscfg->srcLoc[ns][0];
  y = syscfg->srcLoc[ns][1];
  z = syscfg->srcLoc[ns][2];

  kx = syscfg->srcDir[ns][0];
  ky = syscfg->srcDir[ns][1];
  kz = syscfg->srcDir[ns][2];

  /* report if direction vector wrong length... */

#if (PARANOID)
  assert((kx >= -1) && (kx <= 1));
  assert((ky >= -1) && (ky <= 1));
  assert((kz >= -1) && (kz <= 1));
  assert(SQR(kx) + SQR(ky) + SQR(kz) > 0.9999);
  assert(SQR(kx) + SQR(ky) + SQR(kz) < 1.0001);
#endif

  /* Allow for finite source radius */

  if (syscfg->srcRad[ns] > 0)
    {
      double dx, dy, dz, mag;
      double nx, ny, nz, nrm;

    vparallel:
      do
	{
	  /* Generate on a cube then reject based on a sphere */
	  dx = 2 * ran(0) - 1;
	  dy = 2 * ran(0) - 1;
          dz = 2 * ran(0) - 1;
	}
      while (SQR(dx) + SQR(dy) + SQR(dz) > 1.0);

      /* Motion must be in plane perpendicular to initial photon
         direction.  Compute this as the cross-product of my vector
         with the initial k-vector, then rescale back to a unit vector
         to maintain the correct angular distribution */

      nx = ky * dz - kz * dy;
      ny = kz * dx - kx * dz;
      nz = kx * dy - ky * dx;

      if ((nrm = SQR(nx) + SQR(ny) + SQR(nz)) == 0)
        goto vparallel;    /* I know, I know... */
      else
        nrm = sqrt(nrm);   /* Magnitude of the vector */

      /* Turn cross-product into a unit vector to set the angle, and
       * then multiply by a radius selected from a uniform
       * distribution 
       */

      do
	{
	  mag = ran(0);
	}
      while (mag <= 0);

      nx *= mag / nrm;
      ny *= mag / nrm;
      nz *= mag / nrm;

      /* Scale up from a unity-scale vector to the user-specified radius */

      x += nx * (syscfg->srcRad[ns]);
      y += ny * (syscfg->srcRad[ns]);
      z += nz * (syscfg->srcRad[ns]);
    }

  /* Allow for finite source NA */

  if (syscfg->srcNA[ns] > 0)
    {
      double th0, dth;
      double ph0, dph;
      double p1, p2, p3;

      assert(syscfg->srcNA[ns] > 0.0 && syscfg->srcNA[ns] <= 1.0);

      /* The original input direction can be definied as
       * the vector (0,0,1) times a pair of rotation matrices
       * characterized by these two angles */

      th0 = acos(kz);
      ph0 = atan2(ky, kx);

      /* Perturbation d\Omega to the propagation direction */
      /*   NA == n sin(th_0) -> th_0 = asin(NA) (n = 1)    */

      dph = 2 * PI * ran(0);

      do 
        {
          /* Actually sin(theta), convert to sine after we pass
             through fiberNA.  This is not a particularly efficient
             way to select photon angles, but I only have to do it at
             the start of the run. */

          dth = 2*ran(0) - 1;
        }
      while (fiberNA(syscfg->srcNA[ns], dth) != 1);

      dth = asin(dth);

      /* Convert perturbation into rectangular coordinates */

      p1 = sin(dth) * cos(dph);
      p2 = sin(dth) * sin(dph);
      p3 = cos(dth);

      /* Rotate perturbed (0,0,1) vector into its original
       * orientation using the same rotation matrices we would have
       * used for the original (a rotation by th about the y-axis,
       * followed by rotation by ph about the z-axis) 
       */

      kx = cos(th0) * p1 + 0 * p2 + sin(th0) * p3;
      ky = 0 * p1 + 1 * p2 + 0 * p3;
      kz = -sin(th0) * p1 + 0 * p2 + cos(th0) * p3;

      p1 = kx;		  /* Copy back to save a bit of storage space */
      p2 = ky;
      p3 = kz;

      kx = cos(ph0) * p1 + sin(ph0) * p2 + 0 * p3;
      ky = -sin(ph0) * p1 + cos(ph0) * p2 + 0 * p3;
      kz = 0 * p1 + 0 * p2 + 1 * p3;
    }

  /* check rotated length... */

#if (PARANOID)
  assert((kx >= -1) && (kx <= 1));
  assert((ky >= -1) && (ky <= 1));
  assert((kz >= -1) && (kz <= 1));
  assert(SQR(kx) + SQR(ky) + SQR(kz) > 0.9999);
  assert(SQR(kx) + SQR(ky) + SQR(kz) < 1.0001);
#endif

  /* Go from floating point to "fixed" point system */

  PACKCOORD(R, x, y, z);
  renorm(&R);

  /* Get the initial optical properties */

  tissueIndex = TISSUE_TYPE_C(R, syscfg, tissueType);

#if (PARANOID)
  /* make sure photon doesn't start in air... */
  assert(tissueIndex != 0);
  assert(tissueIndex <= syscfg->Ntissue);
#endif

  /* Loop over scattering events until time exceeds the width of the gate
   * or the photon escapes (leaves whole volume or finds itself in air)
   */

  /* Tresid is the time left until I sample the photon distribution again.
   * The interval between samples is fixed (currently, at compile time).
   * Don't collect before the user-specified starting time. */

  Tresid = MAX(syscfg->Tmin, 0.0) + T_SAMPLE;

  while (Ttot < syscfg->Tmax && IN_TISSUE(tissueIndex) &&
         IN_VOLUME_C(R, syscfg))
    {
      /* Calculate new scattering length ---------------------------- */

      do
	{
	  /* rnm must be strictly positive or the log will blow up and
	   * generate a floating-point error on me 
	   */

	  rnm = ran(0);
	}
      while (rnm <= 0.0);

      /* PROPAGATE THE PHOTON --------------------------------------- */

      /* Lresid is an exponentially distributed random number with a mean
       * of 1.0.  It is also is the dimensionless quantity mu*x 
       *
       * Add to previous value to deal with numerical slop. 
       */

      /* I can probably safely ignore nm-scale differences */

      if (Lresid < L_TOL);
        Lresid = -log(rnm);

      while (Lresid >= L_TOL)
	{
	  assert(IN_VOLUME_C(R, syscfg));

	  /* Compute length of this step ---------------------------- */

	  stepT = Tresid * (syscfg->v0 / syscfg->tn[tissueIndex]);
#if 0
	  stepL = Lresid / syscfg->tmus[tissueIndex];
#else
	  if (syscfg->tmus[tissueIndex] > 0)
	    stepL = Lresid / syscfg->tmus[tissueIndex];
	  else
	    stepL = 1e35;	/* Anything larger than system size works */
#endif

	  assert(stepT > 0);
	  assert(stepL > 0);

	  /* Go with whichever limit is smaller */
	  maxstep = MIN(stepT, stepL);

	  /* If this gets bigger than a voxel, it becomes possible to
	     "tunnel" throuh tissue types, which is bad.  Limiting maxstep
	     reduces these problems. */

	  maxstep = MIN(maxstep, 0.5);	/* Half a voxel steps */
          assert(maxstep > 0);

	  /* Do greedy update of the position until something
	     interesting happens.  Since the calculations involved are
	     minimal, we can use relatively small steps through the voxel */

#if (PARANOID)
	  assert(SQR(kx) + SQR(ky) + SQR(kz) != 0);
#endif

          step = 0;

          do
	    {
              /* L_SAMPLE defined in tMCimg.h, L_SAMPLE=0.5 voxel currently */
              step = MIN(step + L_SAMPLE, maxstep);

              Rnew.ix = R.ix;		/* Copy over interger part */
              Rnew.iy = R.iy;
              Rnew.iz = R.iz;

              Rnew.fx = R.fx + kx * step;	/* update fractional */
              Rnew.fy = R.fy + ky * step;
              Rnew.fz = R.fz + kz * step;

              renorm(&Rnew);		/* Renormalize values */

              if (!IN_VOLUME_C(Rnew, syscfg))
                break;

              if (TISSUE_TYPE_C(Rnew, syscfg, tissueType) != tissueIndex)
                break;
	    }
          while (step < maxstep);

	  /* We already know that this step was too big */
	  maxstep = step;

#if (PARANOID)
          assert((step <= maxstep) && (step > 0));
#endif

	  /* Decide what to do with this step ----------------------- */

	  if (!IN_VOLUME_C(Rnew, syscfg))
	    {
	      /* Stepped out of system, photon is officially gone. */

              memcpy(&R, &Rnew, sizeof(MCCOORD));
	    }
          else 
            {
              /* Getting tissue types is expensive.  Find them once
               * and store the results to be re-used below. */

              int oldIdx = tissueIndex;
              int newIdx = TISSUE_TYPE_C(Rnew, syscfg, tissueType);

#if (PARANOID)
              assert((oldIdx >= 0) && (newIdx >= 0));
#endif

              if (oldIdx == newIdx)
                {
                  /* Trivial case, didn't pass an interface */

                  memcpy(&R, &Rnew, sizeof(MCCOORD));
                }
              else if ((syscfg->tn[oldIdx] == syscfg->tn[newIdx]) &&
                       ((syscfg->flags.mirror == 0) || (newIdx != 0)))
                {
                  /* Mostly trivial case, same index of refraction
                   * before and after stepping, so no possibility for
                   * reflection.  Always accept these steps.
                   *
                   * Checks to avoid special case of mirrored BC's and
                   * the final tissue type of zero--this special case
                   * should fall through into the main reflection code
                   * below.  */

                  memcpy(&R, &Rnew, sizeof(MCCOORD));
                }
              else
                {
                  MCCOORD  Rold;

                  /* Hit a tissue interface, reflect (or not) ----------- */

#if (PARANOID)
                  /* K is supposed to be a unit vector! */
                  assert((SQR(kx)+SQR(ky)+SQR(kz) > 0.9999) &&
                         (SQR(kx)+SQR(ky)+SQR(kz) < 1.00001));

                  assert(oldIdx != newIdx);
                  assert(syscfg->flags.mirror ||
                         (syscfg->tn[oldIdx] != syscfg->tn[newIdx]));
#endif

                  /* Track back to boundary to keep book-keeping straight */

                  memcpy(&Rold, &Rnew, sizeof(MCCOORD));

                  find_interface(&R, &Rnew, syscfg, tissueType);

                  /* Update step to reflect change in location */

                  step -= DISTANCE(Rnew, Rold);

#if (PARANOID)
                  assert(TISSUE_TYPE_C(Rnew, 
                                       syscfg, tissueType) != tissueIndex);

                  assert(IN_VOLUME_C(Rnew, syscfg));
                  /*
                  if (syscfg->flags.mirror)
                    assert(TISSUE_TYPE_C(Rnew, syscfg, tissueType) != 0);
                  */
#endif

                  /* reflect_photon() handles mirrored BC's for me */

                  renorm(&R);
                  renorm(&Rnew);

                  if (reflect_photon(syscfg, &R, &Rnew,
                                     &kx, &ky, &kz, seed, tissueType))
                    {
                      /* Photon reflected, k[xyz] already updated */

#if (PARANOID)
                      assert(TISSUE_TYPE_C(R, syscfg, tissueType)     != 0);
#endif

                      memcpy(&Rold, &Rnew, sizeof(MCCOORD));
                      memcpy(&Rnew, &R,    sizeof(MCCOORD));

                      /* Find the other side of the interface */

                      find_interface(&Rold, &Rnew, syscfg, tissueType);

                      /* Update step to reflect change in location */

                      step -= DISTANCE(Rnew, Rold);

                      /* Now on correct side */
                      memcpy(&R, &Rnew, sizeof(MCCOORD));

#if (PARANOID)
                      assert(TISSUE_TYPE_C(R, syscfg, tissueType) == oldIdx);
#endif
                    }
                  else
                    {
                      /* Photon refracted, k[xyz] already updated ------- */

                      /* Already on correct side */
                      memcpy(&R, &Rnew, sizeof(MCCOORD));
                    }
                }
	    }

#if (PARANOID)
	  assert(IN_TISSUE(tissueIndex));
#endif

	  /* Update based on actual step taken: time ---------------- */

	  Ttot   += step / (syscfg->v0 / syscfg->tn[tissueIndex]);
	  Tresid -= step / (syscfg->v0 / syscfg->tn[tissueIndex]);

	  /* Update based on actual step taken: distance ------------ */

	  Lresid -= step * syscfg->tmus[tissueIndex];

	  lenTiss[tissueIndex] += (float)step;

	  /* If Tresid is zero, score the photon */

	  if (Tresid < T_TOL)
	    {
	      /* record photon in 2pt array, unless out of time */

	      if (Ttot < syscfg->Tmax)
		{
		  /* Compute new amplitude [and phase] from updated lenTiss */
		  rotate_phase(&P2pt, &Q2pt, syscfg, lenTiss);

                  renorm(&R); /* Just in case */

		  score_photon(syscfg, tissueType,
                               &R, Ttot, P2pt, Q2pt, II, JJ);
		}

	      /* Reset Tresid to sample interval, subtract off the
	       * current value just to deal with any minor numerical fuzz.
	       */

	      Tresid = T_SAMPLE - Tresid;
	      assert(Tresid > 0);
	    }

	  /* Update tissue type after making the step */

	  tissueIndex = TISSUE_TYPE_C(R, syscfg, tissueType);

#if (PARANOID)
	  assert(tissueIndex <= syscfg->Ntissue);
#endif

	  /* ----------------------------------------------------------
	   * Break out of the inner loop if we're no longer able to
	   * propagate the photon
	   */

	  if (Ttot < syscfg->Tmax)
            {
#if (DEBUG) && (0)
              printf("Photon out of time\n");
#endif
              break;
            }

	  if (! IN_VOLUME_C(R, syscfg))
            {
#if (DEBUG)
              printf("Photon left imaging volume\n");
#endif
              break;
            }

	  if (! IN_TISSUE(tissueIndex))
            {
#if (DEBUG)
              printf("Photon no longer in tissue\n");
#endif
              break;
            }
	}		      /* Propagate photon */

      /* --------------------------------------------------------------
       * Scatter into new direction using a Heyney-Greenstein
       * distribution if Lresid is zero. 
       */

      if (IN_VOLUME_C(R,syscfg))
        if ((Lresid < L_TOL) && IN_TISSUE(tissueIndex))
          scatter(syscfg->tg[tissueIndex], &kx, &ky, &kz, seed);
    }			      /* Loop until end of single photon */

  /* Score the exiting photon --------------------------------------- */

  if (IN_VOLUME_C(R, syscfg)     && (Ttot < syscfg->Tmax))
    if (!IN_TISSUE(tissueIndex))
      {
	/* Compute new amplitude [and phase] from updated lenTiss */
	rotate_phase(&P2pt, &Q2pt, syscfg, lenTiss);

	/* Save exiting photon to 2pt tables (exit with negative fluence) */

        renorm(&R); /* Just in case */
	score_photon(syscfg, tissueType, &R, Ttot, -P2pt, -Q2pt, II, JJ);

        UNPACKCOORD(R, x, y, z);

        /* Write out photon to .his file.   Parallel and single-threaded
         * versions use different savehis() functions, so call an external
         * wrapper routine here */

        write_photon(syscfg, fp, x, y, z, kx, ky, kz, Ttot, lenTiss);
      }

  return;
}

/* ********************************************************************
 * Tissue changed, have the potential for reflection.
 *
 * Return 1 for reflection, 0 otherwise
 */

static int reflect_photon(const struct Config *syscfg,
                          const MCCOORD *Rold, const MCCOORD *Rnew,
                          double *kx, double *ky, double *kz,
                          int *seed, const TISSUE ***tissueType)
{
  double R_fresnel, R_fresnel_1, R_fresnel_2;
  double ctheta, ctheta_out, ctheta_c2;
  double sn1, sn2, sn3;
  double nold, nnew, rnm;
  int    norm, tissueIndex_old, tissueIndex_new;

#if (PARANOID)
  assert(IN_VOLUME_C(*Rnew, syscfg));
#endif

  tissueIndex_old = TISSUE_TYPE_C(*Rold, syscfg, tissueType);
  tissueIndex_new = TISSUE_TYPE_C(*Rnew, syscfg, tissueType);

  assert(tissueIndex_old != tissueIndex_new);

#if (PARANOID)
  assert(tissueIndex_old >= 0 && tissueIndex_old <= syscfg->Ntissue);
  assert(tissueIndex_new >= 0 && tissueIndex_new <= syscfg->Ntissue);
#endif

  nold = syscfg->tn[tissueIndex_old];
  nnew = syscfg->tn[tissueIndex_new];

  /* Change place-holder values into 1.0 */

  if (nold == IDX_AIR)
    nold = 1.0;
  if (nnew == IDX_AIR)
    nnew = 1.0;

  /* 2.75 is physically reasonable, but arbitrary */
#if (PARANOID)
  assert((nold >= 1.0) && (nold < 2.75));
  assert((nnew >= 1.0) && (nnew < 2.75));
#endif

  /* CALCULATE SURFACE NORMAL VECTOR OUT OF TISSUE */

  /* There are really only 3 possible cases using
   * our linearized approach
   *
   * 1) Along a single coordinate axis (cube face)
   * 2) Pair of coordinate axis (cube edge)
   * 3) Along 3 coordinate axes (cube corner)
   * 
   * Take the interface to be orthogonal to one
   * of these three instances 
   */

  /* THIS IS SERIOUSLY BROKEN! REPORTS CORNERS IF WE CROSS DIAGONALLY
     INTO AN ADJACENT VOXEL, EVEN THOUGH THE INTERFACE IS PLANAR! */

  sn1 = Rnew->ix - Rold->ix;
  sn2 = Rnew->iy - Rold->iy;
  sn3 = Rnew->iz - Rold->iz;

  /* inew-i = 0 or +/- 1, so the square is always 0 or 1 */
  norm = SQR(Rnew->ix - Rold->ix) +  
    SQR(Rnew->iy - Rold->iy) + SQR(Rnew->iz - Rold->iz);

  /* Because its computed from integers, the normalization
   * factors can be specified excatly in advance - saves time */

  switch (norm)
    {
      case 1:
	break;			  /* Already normalized */
      case 2:
	sn1 /= SQRT2;
	sn2 /= SQRT2;
	sn3 /= SQRT2;
	break;
      case 3:
	sn1 /= SQRT3;
	sn2 /= SQRT3;
	sn3 /= SQRT3;
	break;
      default:
	fprintf(stderr, "Program bug, norm = %d != 1,2, or 3\n", norm);
#if (PARANOID)
	assert((norm == 1) || (norm == 2) || (norm == 3));
#endif
	exit(1);
    }

  /* cosine of angle of incidence */
  ctheta = (*kx) * sn1 + (*ky) * sn2 + (*kz) * sn3;

  /* squared cosine of critical angle (incident) */
  ctheta_c2 = 1 - SQR(nnew) / SQR(nold);

  if (SQR(ctheta) < ctheta_c2)
    {
      /* Total internal reflection, always reflected */

      R_fresnel = 1;
      ctheta_out = -999;	/* dummy value */
    }
  else if (syscfg->flags.mirror &&
	   (TISSUE_TYPE_C(*Rnew, syscfg, tissueType) == 0))
    {
      /* Perfectly reflecting air-tissue bound's, always reflected */

      R_fresnel = 1;
      ctheta_out = -999;	/* dummy value */
    }
  else
    {
      /* cosine of transmitted angle */

      ctheta_out = sqrt(1 - (SQR(nold) / SQR(nnew)) * (1 - SQR(ctheta)));

      /* Calculate Fresnel reflection coefficient */

      R_fresnel_1 = ((nold * ctheta_out - nnew * ctheta) /
		     (nold * ctheta_out + nnew * ctheta));
      R_fresnel_2 = ((nold * ctheta - nnew * ctheta_out) /
		     (nold * ctheta + nnew * ctheta_out));

      R_fresnel = (SQR(R_fresnel_1) + SQR(R_fresnel_2)) / 2;
    }

  /* Reflect if ran() < R_fresnel */

  if ((rnm = ran(0)) <= R_fresnel)
    {
      /* Reflection occurred, adjust wave-vector, k.n = -k.n */

#if (PARANOID)
      assert(TISSUE_TYPE_C(*Rold, syscfg, tissueType) != 0);
#endif

      *kx -= 2 * ctheta * sn1;
      *ky -= 2 * ctheta * sn2;
      *kz -= 2 * ctheta * sn3;

#if (PARANOID)
      assert((*kx >= -1) && (*kx <= 1));
      assert((*ky >= -1) && (*ky <= 1));
      assert((*kz >= -1) && (*kz <= 1));
      assert(SQR(*kx) + SQR(*ky) + SQR(*kz) > 0.9999);
      assert(SQR(*kx) + SQR(*ky) + SQR(*kz) < 1.0001);
#endif

      return 1;
    }
  else
    {
      double knorm;

      /* Transmission occured, refract according to Snell's law */

#if (PARANOID)
      assert((ctheta_out >= -1) && (ctheta_out <= 1));
#endif

      /* Subtract cos(theta), leaving sin(theta) terms */
      *kx -= ctheta * sn1;
      *ky -= ctheta * sn2;
      *kz -= ctheta * sn3;

      /* Rescale sin(theta), preserving sin(phi),cos(phi)
       * to get sin(theta') [E_{||} terms] */

      *kx *= nold / nnew;
      *ky *= nold / nnew;
      *kz *= nold / nnew;

      /* Add back in cos(theta) [E_\perp terms] */

      *kx += ctheta_out * sn1;
      *ky += ctheta_out * sn2;
      *kz += ctheta_out * sn3;

#if (PARANOID)
      assert((*kx >= -1) && (*kx <= 1));
      assert((*ky >= -1) && (*ky <= 1));
      assert((*kz >= -1) && (*kz <= 1));
#endif

      /* Re-normalize to keep K a unit vector */

      knorm = sqrt(SQR(*kx) + SQR(*ky) + SQR(*kz));
      *kx /= knorm;
      *ky /= knorm;
      *kz /= knorm;

#if (PARANOID)
      assert(SQR(*kx) + SQR(*ky) + SQR(*kz) > 0.9999);
      assert(SQR(*kx) + SQR(*ky) + SQR(*kz) < 1.0001);
#endif
    }

  return 0;
}

/**************************************************************
  HANDLE PHASE ROTATION BETWEEN REAL AND IMAGINARY PART
**************************************************************/

/* Originally, this handled just the increment due to the current
 *  step.  That proved sufficiently non-robust against round-off
 *  errors to warrent recomputing the photon amplitude and phase
 *  every step of the simulation.  Since I was already computing
 *  an exp(), a sin(), and a cos(), this is not significantly
 *  slower than the old way, which is also nice.
 */

static void rotate_phase(TWOPT_T *u, TWOPT_T *v, 
                         const struct Config *syscfg, const float *lenTiss)
{
  double len, phi, mag;
  int i;

  len = 0.0;
  phi = 0.0;

  for (i = 1; i <= syscfg->Ntissue; i++)
    {
      len += lenTiss[i] * syscfg->tmua[i];		/* \mu l */
      phi += lenTiss[i] * (syscfg->tn[i] * syscfg->k0);	/* phase */
    }

  mag = exp(-len);

#if (PARANOID)
  assert(mag > 0);
#endif

  if (syscfg->frequency == 0.0)		/* CW case */
    {
      *u = mag;
      *v = 0;
    }
  else
    /* RF case */
    {
      *u = mag * cos(phi);
      *v = mag * sin(phi);
    }

  return;
}

/**************************************************************
 Renormalized coordinte structure to put fractional part back
  on to 0 <= f < 1.
**************************************************************/

static void renorm(MCCOORD *r)
{
  /* Move the fractional portions back to 0 <= f < 1. Adjust integer 
   * portions to match
   */

  if ((r->fx < 0.0) || (r->fx >= 1.0))
    {
      for ( ; r->fx  < 0.0; r->fx += 1.0) r->ix -= 1;
      for ( ; r->fx >= 1.0; r->fx -= 1.0) r->ix += 1;
    }

  if ((r->fy < 0.0) || (r->fy >= 1.0))
    {
      for ( ; r->fy  < 0.0; r->fy += 1.0) r->iy -= 1;
      for ( ; r->fy >= 1.0; r->fy -= 1.0) r->iy += 1;
    }

  if ((r->fz < 0.0) || (r->fz >= 1.0))
    {
      for ( ; r->fz  < 0.0; r->fz += 1.0) r->iz -= 1;
      for ( ; r->fz >= 1.0; r->fz -= 1.0) r->iz += 1;
    }

  return;
}

/**************************************************************
 *
 * CALCULATE THE NEW SCATTERING ANGLE USING HENYEY-GREENSTEIN
 *
 *************************************************************/

static void scatter(double gg, double *kx, double *ky, double *kz, int *seed)
{
#if (PARANOID)
  /* Make sure the anisotropy is physically meaningful */

  assert((gg >= -0.5) && (gg <= 1.0));

  /* Sanity-check the input direction cosines */

  assert((*kx >= -1) && (*kx <= 1));
  assert((*ky >= -1) && (*ky <= 1));
  assert((*kz >= -1) && (*kz <= 1));
  assert(SQR(*kx) + SQR(*ky) + SQR(*kz) > 0.9999);
  assert(SQR(*kx) + SQR(*ky) + SQR(*kz) < 1.0001);
#endif

  /* If gg is identically zero, the calculated distribution has a
   * divide-by-zero, but gg==0 is just the isotropic case, which is
   * trivial to handle */

  if (gg == 0.0)
    {
      double nrm, C1, C2, C3;

      /* Select in a cube, reject if it falls outside the sphere.
         Gives us a vector uniformly distributed through-out the
         volume of the sphere.  Rescale to unit length after.  May not
         be the most efficient algorithm imaginable, but it does
         guarantee uniform angular sampling. */

      do
	{
	  C1 = 2 * ran(0) - 1;
	  C2 = 2 * ran(0) - 1;
	  C3 = 2 * ran(0) - 1;
	}
      while (SQR(C1) + SQR(C2) + SQR(C3) > 1.0);

      /* (kx,ky,kz) must be a unit vector, rescale vector length */
      nrm = sqrt(SQR(C1) + SQR(C2) + SQR(C3));

#if (PARANOID)
      assert(nrm > 0.0);
#endif

      *kx = C1 / nrm;
      *ky = C2 / nrm;
      *kz = C3 / nrm;
    }
  else
    {			      /* gg != 0 */
      double kxo, kyo, kzo;   /* OLD DIRECTION COSINES */
      double phi, cphi, sphi, theta, stheta, ctheta;
      double rnm, nrm, foo;

      kxo = *kx;
      kyo = *ky;
      kzo = *kz;

      phi = 2 * PI * ran(0);  /* Scatter randomly in transverse plane */
      cphi = cos(phi);
      sphi = sin(phi);

      rnm = ran(0);	      /* Scatter with H-G dist in prop. dir */

      foo = (1.0 - SQR(gg)) / (1.0 - gg + 2.0 * gg * rnm);
      foo = SQR(foo);
      ctheta = (1.0 + SQR(gg) - foo) / (2.0 * gg);

#if (PARANOID)
      assert((ctheta >= -1) && (ctheta <= 1));
#endif

      theta = acos(ctheta);
      stheta = sin(theta);

      if ((*kz > -1) && (*kz < 1))
	{
	  double bar = sqrt(1 - SQR(kzo));
#if (PARANOID)
	  assert((bar > 0) && (bar <= 1.0));
#endif

	  /* Values of bar on the order 1e-6 causes real problems.
	   * Multiply actual equation by bar and renormalize to 1.0
	   * at the end. */

	  *kx = stheta * (kxo * kzo * cphi - kyo * sphi) + ctheta * kxo * bar;
	  *ky = stheta * (kyo * kzo * cphi + kxo * sphi) + ctheta * kyo * bar;
	  *kz = -stheta * cphi * SQR(bar) + ctheta * kzo * bar;

	  nrm = sqrt(SQR(*kx) + SQR(*ky) + SQR(*kz));

	  *kx /= nrm;
	  *ky /= nrm;
	  *kz /= nrm;
	}
      else
	{
	  /* Vertical direction causes problems, treat as a special case */

	  *kx = stheta * cphi;
	  *ky = stheta * sphi;
	  *kz = ctheta * kzo;
	}
    }

#if (PARANOID)
  /* Sanity-check the output direction cosines */

  assert((*kx >= -1) && (*kx <= 1));
  assert((*ky >= -1) && (*ky <= 1));
  assert((*kz >= -1) && (*kz <= 1));

  assert(SQR(*kx) + SQR(*ky) + SQR(*kz) > 0.9999);
  assert(SQR(*kx) + SQR(*ky) + SQR(*kz) < 1.0001);
#endif

  return;
}

/* ********************************************************************
 *
 * Record a photon in the 2-pt table
 *
 * ********************************************************************/

static void score_photon(const struct Config *syscfg,
			 const TISSUE ***tissueType, const MCCOORD *R,
			 double Ttot, TWOPT_T P2pt, TWOPT_T Q2pt,
			 TWOPT_T *II, TWOPT_T *JJ)
{
  int tindex, VoxN;

  if (syscfg->flags.gen_twopt)
    {
      if (IN_ROI_FULL_C(*R, Ttot, syscfg))
	{
	  tindex = (int)((Ttot - syscfg->Tmin) / syscfg->stepT);

#if (PARANOID)
	  assert(tindex < syscfg->nTstep);
#endif

	  /* Compute index into tables */

	  if (syscfg->flags.matlab_order)
	    {
	      /* Use Matlab ordering so I can dump to disk 
	       * the entire block of memory later 
	       */

	      /* Casting to int is an expensive operation, leave as floats as
	       * long as possible and cast only at the end.  Add a tiny bit
	       * before the cast to avoid the inevitable round-off problems */

	      VoxN = (R->iy - syscfg->Iymin) +
                      syscfg->nIystep * ((R->ix - syscfg->Ixmin) +
                      syscfg->nIxstep * ((R->iz - syscfg->Izmin) +
                      syscfg->nIzstep * (tindex)));
	    }
	  else
	    {
	      /* Use normal C ordering */

	      /* Casting to int is an expensive operation, leave as floats as
	       * long as possible and cast only at the end.  Add a tiny bit 
	       * before the cast to avoid the inevitable round-off problems */

	      VoxN = (R->ix - syscfg->Ixmin) +
                      syscfg->nIxstep * ((R->iy - syscfg->Iymin) +
                      syscfg->nIystep * ((R->iz - syscfg->Izmin) +
                      syscfg->nIzstep * (tindex)));
	    }

#if (PARANOID)
	  assert(VoxN < (syscfg->nIxstep * syscfg->nIystep *
			 syscfg->nIzstep * syscfg->nTstep));
          assert(VoxN >= 0);
#endif

	  if (II != NULL)
	    II[VoxN] += P2pt;

	  if (JJ != NULL)
	    JJ[VoxN] += Q2pt;
	}
    }

  return;
}

/***********************************************************************
 * Advance to the interface.  The interface is on the line segment
 * [x0,y0,z0] -- [x1,y1,z1].  Motion is along the line segment.
 *
 * The algorithm promises to go to an interface.  It does NOT promise
 * to find the _first_ interface.
 **********************************************************************/

static void find_interface(const MCCOORD *_R0, MCCOORD *_R1,
                           const struct Config *system, 
                           const TISSUE ***tissueType)
{
  MCCOORD R, R0, R1;
  double sep;
  TISSUE ti0, ti1, ti;

  /* Make private (modifiable) copies */

  memcpy(&R0, _R0, sizeof(MCCOORD));
  memcpy(&R1, _R1, sizeof(MCCOORD));

  sep = 0;

  /* It doesn't matter what these are, so long as they're different */

  ti0 = TISSUE_TYPE_C(R0, system, tissueType);
  ti1 = TISSUE_TYPE_C(R1, system, tissueType);

  do
    {
      /* Find the midpoint between R0 and R1 */

      assert(ti0 != ti1);

      R.ix = R0.ix;
      R.fx = (R1.ix - R0.ix)/2.0 + (R0.fx + R1.fx) / 2.0;

      R.iy = R0.iy;
      R.fy = (R1.iy - R0.iy)/2.0 + (R0.fy + R1.fy) / 2.0;

      R.iz = R0.iz;
      R.fz = (R1.iz - R0.iz)/2.0 + (R0.fz + R1.fz) / 2.0;

      /* Re-balance fractional portions */

      renorm(&R);

      ti = TISSUE_TYPE_C(R, system, tissueType);

      if (ti == ti0)
        {
          memcpy(&R0, &R, sizeof(MCCOORD));
          ti0 = ti;
        }
      else
        {
          /* Accept, even if ti != ti1 */

          memcpy(&R1, &R, sizeof(MCCOORD));
          ti1 = ti;
        }

      assert(ti0 != ti1);

      sep = DISTANCE(R0,R1);
    }
  while (sep > 1e-6);

  /* Copy back */

  renorm(&R1);
  memcpy(_R1, &R1, sizeof(MCCOORD));

  return;
}

/* Given a fiber NA and a trial photon angle sin(th) (0 is straight
   ahead), accept (returns 1) or reject (returns 0) the angle based on
   a selection from a Gaussian distribution with FWHM of 2*NA. */

static int fiberNA(double NA, double sinth)
{
  double a, b;

#if (HAVE_ASSERT)
  assert((sinth >= -1) && (sinth <= 1));
#endif

  if (NA <= 0.0 || NA > 1)
    return 0;           /* Delta-function distributions, will never match */

  /* Normalization for distribution */

  b = SQR(NA) / LN2;
  a = sqrt(b/PI) * ERF(1/sqrt(b));

  /* Accept if p < f(x) */

  if (ran(0) < exp(-b * SQR(sinth)) / a)
    return 1;
  else
    return 0;
}
